from typing import Tuple, Iterator, Optional, List, Callable, Union

import torch
import torch.nn.functional as F
from torch import Tensor
from torch.nn import Module, Parameter

__all__ = ['Readout', 'fit_readout', 'fit_and_validate_readout']


class Readout(Module):
    """
    A linear readout

    Linear model with bias :math:`y = W x + b`, like Linear in Torch.
    """
    weight: Parameter  # (targets × features)
    bias: Parameter  # (targets)

    def __init__(self, num_features: int, num_targets: int):
        """
        New readout

        :param num_features: Number of input features
        :param num_targets: Number of output targets
        """
        super().__init__()
        self.weight = Parameter(Tensor(num_targets, num_features), requires_grad=False)
        self.bias = Parameter(Tensor(num_targets), requires_grad=False)

    def forward(self, x: Tensor) -> Tensor:
        return F.linear(x, self.weight, self.bias)

    def fit(self, data: Union[Iterator[Tuple[Tensor, Tensor]], Tuple[Tensor, Tensor]],
            regularization: Union[Optional[float], List[float]] = None,
            validate: Optional[Callable[[Tuple[Tensor, Tensor]], float]] = None):
        """
        Fit readout to data

        :param data: Dataset of (features, targets) tuples, or single pair
        :param regularization: Ridge regression lambda, or lambda if validation requested
        :param validate: Validation function, if regularization is to be selected
        """
        if not hasattr(data, '__next__'):
            data = iter([data])
        if callable(validate):
            self.weight.data, self.bias.data = fit_and_validate_readout(data, regularization, validate)
        else:
            self.weight.data, self.bias.data = fit_readout(data, regularization)

    @property
    def num_features(self) -> int:
        """
        Input features

        :return: Number of input features
        """
        return self.weight.shape[1]

    @property
    def num_targets(self) -> int:
        """
        Output targets

        :return: Number of output targets
        """
        return self.weight.shape[0]

    def __repr__(self):
        return f'Readout(features={self.num_features}, targets={self.num_targets})'


def fit_readout(data: Iterator[Tuple[Tensor, Tensor]], regularization: Optional[float] = None) -> Tuple[Tensor, Tensor]:
    """
    Ridge regression for big data

    Fits a linear model :math:`y = W x + b` with regularization.
    See:
    T. Zhang & B. Yang (2017). An exact approach to ridge regression for big data.
    Computational Statistics, 32(3), 909–928. https://doi.org/10.1007/s00180-017-0731-5

    :param data: Batch dataset of pairs (x, y) with samples on rows
    :param regularization: Regularization constant for ridge regression (default null)
    :return: A pair of tensors (W, b)
    """
    # Compute sufficient statistics for regression
    x, y = next(data)
    Syy = y.square().sum(dim=0)  # (targets)
    Sxy = x.t() @ y  # (features × targets)
    Sxx = x.t() @ x  # (features × features)
    Sy = y.sum(dim=0)  # (targets)
    Sx = x.sum(dim=0)  # (features)
    n = float(x.shape[0])  # samples
    for x, y in data:
        Syy += y.square().sum(dim=0)
        Sxy += x.t() @ y
        Sxx += x.t() @ x
        Sy += y.sum(dim=0)
        Sx += x.sum(dim=0)
        n += x.shape[0]
    # Compute ridge matrices
    Vxx = Sxx.diag() - (Sx.square() / n)
    Vyy = Syy - (Sy.square() / n)
    XX = (Sxx - torch.outer(Sx, Sx) / n) / torch.outer(Vxx, Vxx).sqrt()
    Xy = (Sxy - torch.outer(Sx, Sy) / n) / torch.outer(Vxx, Vyy).sqrt()
    if regularization:
        XX += torch.eye(n=XX.shape[0]).to(XX) * regularization
    # Compute weights
    Ws = torch.linalg.solve(XX, Xy)
    W = Ws * torch.sqrt(Vyy.expand_as(Ws) / Vxx.unsqueeze(-1))
    b = (Sy / n) - (Sx / n) @ W
    return W.t(), b


def fit_and_validate_readout(data: Iterator[Tuple[Tensor, Tensor]], regularization_constants: List[float],
                             get_validation_error: Callable[[Tuple[Tensor, Tensor]], float]) -> Tuple[Tensor, Tensor]:
    """
    Ridge regression for big data, with efficient regularization selection

    Fits a linear model :math:`y = W x + b` with regularization.
    See:
    T. Zhang & B. Yang (2017). An exact approach to ridge regression for big data.
    Computational Statistics, 32(3), 909–928. https://doi.org/10.1007/s00180-017-0731-5

    :param data: Batch dataset of pairs (x, y) with samples on rows
    :param regularization_constants: Regularization constants for ridge regression (including none)
    :param get_validation_error: Evaluate validation error for a regression pair (W, b)
    :return: A pair of tensors (W, b)
    """
    # Compute sufficient statistics for regression
    x, y = next(data)
    Syy = y.square().sum(dim=0)  # (targets)
    Sxy = x.t() @ y  # (features × targets)
    Sxx = x.t() @ x  # (features × features)
    Sy = y.sum(dim=0)  # (targets)
    Sx = x.sum(dim=0)  # (features)
    n = float(x.shape[0])  # samples
    for x, y in data:
        Syy += y.square().sum(dim=0)
        Sxy += x.t() @ y
        Sxx += x.t() @ x
        Sy += y.sum(dim=0)
        Sx += x.sum(dim=0)
        n += x.shape[0]
    # Compute ridge matrices
    Vxx = Sxx.diag() - (Sx.square() / n)
    Vyy = Syy - (Sy.square() / n)
    XX = (Sxx - torch.outer(Sx, Sx) / n) / torch.outer(Vxx, Vxx).sqrt()
    Xy = (Sxy - torch.outer(Sx, Sy) / n) / torch.outer(Vxx, Vyy).sqrt()
    # Compute and select weights
    best_validation_error, best_W, best_b = None, None, None
    for regularization in regularization_constants:
        # Compute weights
        XXr = (XX + torch.eye(n=XX.shape[0]).to(XX) * regularization) if regularization else XX
        Ws = torch.linalg.solve(XXr, Xy)
        W = Ws * torch.sqrt(Vyy.expand_as(Ws) / Vxx.unsqueeze(-1))
        b = (Sy / n) - (Sx / n) @ W
        # Validate, select
        validation_error = get_validation_error((W.t(), b))
        if best_validation_error is None or validation_error < best_validation_error:
            best_validation_error, best_W, best_b = validation_error, W.t(), b
        print(f'   {regularization:e}: {validation_error}')
    return best_W, best_b



from typing import Optional

import torch
from torch import Size, Tensor

__all__ = ['uniform', 'normal', 'ring', 'ones', 'rescale_']


def uniform(size: Size, rho: Optional[float] = None, sigma: Optional[float] = None,
            scale: Optional[float] = None) -> Tensor:
    """
    Uniform random tensor

    Can either be rescaled according to spectral radius `rho`, spectral norm `sigma`, or `scale`.

    :param size: Size of tensor
    :param rho: Spectral radius
    :param sigma: Spectral norm
    :param scale: Simple rescaling of the standard random matrix
    :return: A random tensor
    """
    W = torch.empty(size).uniform_(-1, 1)
    rescale_(W, rho, sigma, scale)
    return W.data


def normal(size: Size, rho: Optional[float] = None, sigma: Optional[float] = None,
           scale: Optional[float] = None) -> Tensor:
    """
    Normal random tensor

    Can either be rescaled according to spectral radius `rho`, spectral norm `sigma`, or `scale`.

    :param size: Size of tensor
    :param rho: Spectral radius
    :param sigma: Spectral norm
    :param scale: Simple rescaling of the standard random matrix
    :return: A random tensor
    """
    W = torch.empty(size).normal_(mean=0, std=1)
    rescale_(W, rho, sigma, scale)
    return W.data


def ring(size: Size, rho: Optional[float] = None, sigma: Optional[float] = None,
         scale: Optional[float] = None) -> Tensor:
    """
    Ring matrix

    See:
    C. Gallicchio & A. Micheli (2020). Ring Reservoir Neural Networks for Graphs.
    In 2020 International Joint Conference on Neural Networks (IJCNN), IEEE.
    https://doi.org/10.1109/IJCNN48605.2020.9206723

    :param size: Size of tensor (must be square)
    :param rho: Spectral radius (equivalent to others)
    :param sigma: Spectral norm (equivalent to others)
    :param scale: Simple rescaling of the matrix (equivalent to others)
    :return: A re-scaled ring matrix
    """
    assert (len(size) == 2) and (size[0] == size[1])
    assert any(arg is not None for arg in [rho, sigma, scale])
    if scale is None:
        scale = rho if sigma is None else sigma
    W = torch.eye(size[0]).roll(1, 0) * scale
    return W.data


def ones(size: Size, rho: Optional[float] = None, sigma: Optional[float] = None,
         scale: Optional[float] = None) -> Tensor:
    """
    Ones tensor

    Can either be rescaled according to spectral radius `rho`, spectral norm `sigma`, or `scale`.

    :param size: Size of tensor
    :param rho: Spectral radius
    :param sigma: Spectral norm
    :param scale: Simple rescaling of the standard random matrix
    :return: A random tensor
    """
    W = torch.ones(size)
    rescale_(W, rho, sigma, scale)
    return W.data


def rescale_(W: Tensor, rho: Optional[float] = None, sigma: Optional[float] = None,
             scale: Optional[float] = None) -> Tensor:
    """
    Rescale a matrix in-place

    Can either be rescaled according to spectral radius `rho`, spectral norm `sigma`, or `scale`.

    :param W: Matrix to rescale
    :param rho: Spectral radius
    :param sigma: Spectral norm
    :param scale: Simple rescaling of the standard random matrix
    :return: Rescaled matrix
    """
    if rho is not None:
        return W.div_(torch.linalg.eigvals(W).abs().max()).mul_(rho)
    elif sigma is not None:
        return W.div_(torch.linalg.matrix_norm(W, ord=2)).mul_(sigma)
    elif scale is not None:
        return W.mul_(scale)



import os.path
from typing import Union, Iterable, List, Optional

import numpy as np
import pandas as pd
import torch
from torch_geometric.data import download_url, extract_zip
from torch_geometric_temporal import TwitterTennisDatasetLoader, PedalMeDatasetLoader, WikiMathsDatasetLoader
from torch_geometric_temporal.dataset import ChickenpoxDatasetLoader
from tqdm import tqdm

from graphesn.data import DynamicData, TemporalData

__all__ = ['TGKDataset', 'chickenpox_dataset', 'twitter_tennis_dataset', 'pedalme_dataset', 'wiki_maths_dataset']


class TGKDataset:
    """
    Susceptible-Infected dynamic graphs, from the "Temporal Graph Kernels" paper

    See:
    Oettershagen, Kriege, Morris, Mutzel (2020). Temporal Graph Kernels for Classifying Dissemination Processes.
    Proceedings of the 2020 SIAM International Conference on Data Mining, pp. 496–504.
    Available at http://arxiv.org/abs/1911.05496

    :param name: Dataset name (e.g. 'dblp_ct1')
    :param root: Root directory for data
    """

    def __init__(self, name: str, root: str = '/tmp'):
        self.name = name
        self.root = root
        path = self._check_and_download()
        self.data = self._load_data(path)

    def _check_and_download(self) -> str:
        path = os.path.join(self.root, self.name)
        if not os.path.isdir(path):
            zip_file = download_url(f'https://www.chrsmrrs.com/graphkerneldatasets/{self.name}.zip', self.root)
            extract_zip(zip_file, self.root)
            os.unlink(zip_file)
        return path

    def _load_data(self, path: str) -> List[DynamicData]:
        data = []
        edges = np.loadtxt(os.path.join(path, f'{self.name}_A.txt'), delimiter=',', dtype=np.int64) - 1
        indicator = np.loadtxt(os.path.join(path, f'{self.name}_graph_indicator.txt'), dtype=np.int64) - 1
        timestamps = np.loadtxt(os.path.join(path, f'{self.name}_edge_attributes.txt'), dtype=np.int64)
        y = np.loadtxt(os.path.join(path, f'{self.name}_graph_labels.txt'), dtype=np.int64)
        df = pd.read_csv(os.path.join(path, f'{self.name}_node_labels.txt'), names=['t0', 'x0', 't1', 'x1'],
                         dtype={'t0': np.int64, 'x0': np.float32, 't1': 'Int64', 'x1': np.float32})
        t1, x1 = df.t1.to_numpy(dtype=np.int64, na_value=-1), df.x1.to_numpy(dtype=np.float32)
        T = max(t1.max(), timestamps.max()) + 1
        change_mask = (t1 >= 0)
        x = np.zeros((len(t1), T), dtype=np.float32)
        x[change_mask, t1[change_mask]] = x1[change_mask]
        x = np.cumsum(x, axis=1)
        pos = np.cumsum(np.bincount(indicator + 1))
        empty_adj = torch.empty((2, 0), dtype=torch.int64)
        for sample in tqdm(np.unique(indicator)):
            edge_slice = np.bitwise_and(edges[:, 0] >= pos[sample], edges[:, 0] < pos[sample + 1])
            edge_index = [empty_adj] * T
            for t in np.unique(timestamps[edge_slice]):
                time_edge_slice = np.bitwise_and(edge_slice, timestamps == t)
                edge_index[t] = torch.tensor(edges[time_edge_slice] - pos[sample]).t()
            unmasked = np.unique(
                np.concatenate([timestamps[edge_slice], t1[[sample]]]) if t1[sample] >= 0 else timestamps[edge_slice])
            mask = np.zeros((T, pos[sample + 1] - pos[sample]), dtype=bool)
            mask[unmasked, :] = True
            data.append(DynamicData(static_keys=['y'], edge_index=edge_index, y=torch.tensor(y[sample]),
                                    x=torch.tensor(x[pos[sample]:pos[sample + 1]].T).unsqueeze(-1),
                                    mask=torch.tensor(mask)))
        return data

    def __getitem__(self, item: Union[int, Iterable[int]]):
        if hasattr(item, '__iter__'):
            return [self.data[index] for index in item]
        else:
            return self.data[item]

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return f'{self.name}({len(self.data)})'


def chickenpox_dataset(target_lags: int = 8, feature_lags: bool = True) -> TemporalData:
    """
    Chickenpox temporal graph dataset

    See:
    Chickenpox Cases in Hungary: a Benchmark Dataset for Spatiotemporal Signal Processing with Graph Neural Networks.
    Benedek Rozemberczki, Paul Scherer, Oliver Kiss, Rik Sarkar, Tamas Ferenci (2021).
    https://arxiv.org/abs/2102.08100

    :param target_lags: Prediction target time-steps lag
    :param feature_lags: If true, include lagged features (default true)
    :return: A single temporal graph
    """
    loader = ChickenpoxDatasetLoader()
    dataset = loader.get_dataset(lags=target_lags)
    return TemporalData(temporal_keys=['x', 'y'],
                        edge_index=torch.from_numpy(dataset.edge_index),
                        x=torch.stack([torch.from_numpy(x.astype('float32')) if feature_lags else torch.from_numpy(
                            x[:, 0].astype('float32')).unsqueeze(dim=-1) for x in dataset.features], dim=0),
                        y=torch.stack(
                            [torch.from_numpy(y.astype('float32')).unsqueeze(dim=-1) for y in dataset.targets], dim=0))


def twitter_tennis_dataset(event_id: str = 'rg17', num_nodes: Optional[int] = None, feature_mode: str = 'encoded',
                           target_offset: int = 1) -> DynamicData:
    """
    Twitter tennis dynamic graph dataset

    See:
    F. Béres, R. Pálovics, A. Oláh, et al. Temporal walk based centrality metric for graph streams.
    Applied Network Science 3, 32 (2018). https://doi.org/10.1007/s41109-018-0080-5

    :param event_id: Name of the event, 'rg17' or 'uo17'
    :param num_nodes: Select top nodes (optional, default all)
    :param feature_mode: Can be None for raw features, or 'encoded' for one hot encoding, or 'diagonal' for identity matrix
    :param target_offset: Prediction off-set for node labels (default 1)
    :return: A single dynamic graph
    """
    loader = TwitterTennisDatasetLoader(event_id=event_id, N=num_nodes, feature_mode=feature_mode,
                                        target_offset=target_offset)
    dataset = loader.get_dataset()
    return DynamicData(edge_index=[torch.from_numpy(edge_index) for edge_index in dataset.edge_indices],
                       edge_weight=[torch.from_numpy(edge_weight.astype('float32')) for edge_weight in
                                    dataset.edge_weights],
                       x=torch.stack([torch.from_numpy(x.astype('float32')) for x in dataset.features], dim=0),
                       y=torch.stack(
                           [torch.from_numpy(y.astype('float32')).unsqueeze(dim=-1) for y in dataset.targets], dim=0))


def pedalme_dataset(target_lags: int = 8, feature_lags: bool = True) -> TemporalData:
    """
    PedalMe temporal graph dataset

    See:
    B. Rozemberczki, P. Scherer, Y. He, et al. (2021). PyTorch Geometric Temporal: Spatiotemporal Signal Processing with Neural Machine Learning Models.
    Proceedings of the 30th ACM International Conference on Information & Knowledge Management (CIKM '21), pp. 4564–4573.
    https://doi.org/10.1145/3459637.3482014

    :param target_lags: Prediction target time-steps lag
    :param feature_lags: If true, include lagged features (default true)
    :return: A single temporal graph
    """
    loader = PedalMeDatasetLoader()
    dataset = loader.get_dataset(lags=target_lags)
    return TemporalData(temporal_keys=['x', 'y'],
                        edge_index=torch.from_numpy(dataset.edge_index),
                        edge_weight=torch.from_numpy(dataset.edge_weight.astype('float32')),
                        x=torch.stack([torch.from_numpy(x.astype('float32')) if feature_lags else torch.from_numpy(
                            x[:, 0].astype('float32')).unsqueeze(dim=-1) for x in dataset.features], dim=0),
                        y=torch.stack(
                            [torch.from_numpy(y.astype('float32')).unsqueeze(dim=-1) for y in dataset.targets], dim=0))


def wiki_maths_dataset(target_lags: int = 8, feature_lags: bool = True) -> TemporalData:
    """
    Wiki Maths temporal graph dataset

    See:
    B. Rozemberczki, P. Scherer, Y. He, et al. (2021). PyTorch Geometric Temporal: Spatiotemporal Signal Processing with Neural Machine Learning Models.
    Proceedings of the 30th ACM International Conference on Information & Knowledge Management (CIKM '21), pp. 4564–4573.
    https://doi.org/10.1145/3459637.3482014

    :param target_lags: Prediction target time-steps lag
    :param feature_lags: If true, include lagged features (default true)
    :return: A single temporal graph
    """
    loader = WikiMathsDatasetLoader()
    dataset = loader.get_dataset(lags=target_lags)
    return TemporalData(temporal_keys=['x', 'y'],
                        edge_index=torch.from_numpy(dataset.edge_index),
                        edge_weight=torch.from_numpy(dataset.edge_weight.astype('float32')),
                        x=torch.stack([torch.from_numpy(x.astype('float32')) if feature_lags else torch.from_numpy(
                            x[:, 0].astype('float32')).unsqueeze(dim=-1) for x in dataset.features], dim=0),
                        y=torch.stack(
                            [torch.from_numpy(y.astype('float32')).unsqueeze(dim=-1) for y in dataset.targets], dim=0))



from typing import Optional, Callable, Union, List

import torch
import torch.nn.functional as F
from torch import Tensor, Size
from torch.nn import Module, Parameter, ModuleList
from torch_geometric.nn import MessagePassing
from torch_geometric.typing import Adj, OptTensor
from torch_sparse import matmul, SparseTensor

from graphesn import matrix

__all__ = ['initializer', 'ReservoirConvLayer', 'GraphReservoir', 'StaticGraphReservoir', 'DynamicGraphReservoir']


def initializer(name: str, **options) -> Callable[[Size], Tensor]:
    """
    Gets a random weight initializer

    :param name: Name of the random matrix generator in `graphesn.matrix`
    :param options: Random matrix generator options
    :return: A random weight generator function
    """
    init = getattr(matrix, name)
    return lambda size: init(size, **options)


class ReservoirConvLayer(MessagePassing):
    """
    A Graph ESN convolution layer

    :param input_features: Input dimension
    :param hidden_features: Reservoir dimension
    :param bias: If bias term is present
    :param activation: Activation function (default tanh)
    """
    input_weight: Parameter
    recurrent_weight: Parameter
    leakage: Parameter
    bias: Optional[Parameter]
    activation: Callable[[Tensor], Tensor]

    def __init__(self, input_features: int, hidden_features: int, bias: bool = False,
                 activation: Union[str, Callable[[Tensor], Tensor]] = 'tanh', **kwargs):
        super().__init__(**kwargs)
        self.input_weight = Parameter(torch.empty(hidden_features, input_features), requires_grad=False)
        self.recurrent_weight = Parameter(torch.empty(hidden_features, hidden_features), requires_grad=False)
        self.leakage = Parameter(torch.ones([]), requires_grad=False)
        self.bias = Parameter(torch.empty(hidden_features), requires_grad=False) if bias else None
        self.activation = activation if callable(activation) else getattr(torch, activation)

    def forward(self, edge_index: Adj, input: Tensor, state: Tensor, edge_weight: OptTensor = None):
        neighbour_aggr = self.propagate(edge_index=edge_index, x=state, edge_weight=edge_weight)
        return self.leakage * self.activation(
            F.linear(input, self.input_weight, self.bias) + F.linear(neighbour_aggr, self.recurrent_weight)) \
               + (1 - self.leakage) * state

    def message(self, x_j: Tensor, edge_weight: OptTensor = None) -> Tensor:
        return x_j if edge_weight is None else edge_weight.view(-1, 1) * x_j

    def message_and_aggregate(self, adj_t: SparseTensor, x: Tensor) -> Tensor:
        return matmul(adj_t, x, self.aggr)

    def initialize_parameters(self, recurrent: Callable[[Size], Tensor], input: Callable[[Size], Tensor],
                              bias: Optional[Callable[[Size], Tensor]] = None, leakage: float = 1.0):
        """
        Initialize reservoir weights

        :param recurrent: Random matrix generator for recurrent weight
        :param input: Random matrix generator for input weight
        :param bias: Random matrix generator for bias, if present
        :param leakage: Leakage constant
        """
        self.recurrent_weight.data = recurrent(self.recurrent_weight.shape)
        self.input_weight.data = input(self.input_weight.shape)
        if self.bias is not None:
            self.bias.data = bias(self.bias.shape)
        self.leakage.data = torch.tensor(leakage)

    @property
    def in_features(self) -> int:
        """Input dimension"""
        return self.input_weight.shape[1]

    @property
    def out_features(self) -> int:
        """Reservoir state dimension"""
        return self.input_weight.shape[0]

    def extra_repr(self) -> str:
        return f'in={self.in_features}, out={self.out_features}, bias={self.bias is not None}'


class GraphReservoir(Module):
    """
    Base class for graph reservoirs

    :param num_layers: Reservoir layers
    :param in_features: Size of input
    :param hidden_features: Size of reservoir (i.e. number of hidden units per layer)
    :param bias: Whether bias term is present
    :param kwargs: Other `ReservoirConvLayer` arguments (activation, etc.)
    """
    layers: ModuleList

    def __init__(self, num_layers: int, in_features: int, hidden_features: int, bias: bool = False, **kwargs):
        super().__init__()
        assert num_layers > 0
        self.layers = ModuleList()
        self.layers.append(
            ReservoirConvLayer(input_features=in_features, hidden_features=hidden_features, bias=bias, **kwargs))
        for _ in range(1, num_layers):
            self.layers.append(
                ReservoirConvLayer(input_features=hidden_features, hidden_features=hidden_features, bias=bias,
                                   **kwargs))

    def initialize_parameters(self, recurrent: Callable[[Size], Tensor], input: Callable[[Size], Tensor],
                              bias: Optional[Callable[[Size], Tensor]] = None, leakage: float = 1.0):
        """
        Initialize reservoir weights for all layers

        :param recurrent: Random matrix generator for recurrent weight
        :param input: Random matrix generator for input weight
        :param bias: Random matrix generator for bias, if present
        :param leakage: Leakage constant
        """
        for layer in self.layers:
            layer.initialize_parameters(recurrent=recurrent, input=input, bias=bias, leakage=leakage)

    @property
    def num_layers(self) -> int:
        """Number of reservoir layers"""
        return len(self.layers)

    @property
    def in_features(self) -> int:
        """Input dimension"""
        return self.layers[0].input_weight.shape[1]

    @property
    def out_features(self) -> int:
        """Embedding dimension"""
        raise NotImplementedError()


class StaticGraphReservoir(GraphReservoir):
    """
    Reservoir for static graphs

    :param num_layers: Reservoir layers
    :param in_features: Size of input
    :param hidden_features: Size of reservoir (i.e. number of hidden units per layer)
    :param bias: Whether bias term is present
    :param pooling: Graph pooling function (optional, default no pooling)
    :param fully: Whether to concatenate all layers' encodings, or use just final layer encoding
    :param max_iterations: Maximum number of iterations (optional, default infinity)
    :param epsilon: Convergence condition (default 1e-6)
    """
    pooling: Optional[Callable[[Tensor, Tensor], Tensor]]
    fully: bool
    max_iterations: Optional[int]
    epsilon: float

    def __init__(self, num_layers: int, in_features: int, hidden_features: int, bias: bool = False,
                 pooling: Optional[Callable[[Tensor, Tensor], Tensor]] = None, fully: bool = False,
                 max_iterations: Optional[int] = None, epsilon: float = 1e-6, **kwargs):
        super().__init__(num_layers, in_features, hidden_features, bias, **kwargs)
        self.pooling = pooling
        self.fully = fully
        self.max_iterations = max_iterations
        self.epsilon = epsilon

    def forward(self, edge_index: Adj, input: Tensor, initial_state: Optional[Union[List[Tensor], Tensor]] = None,
                batch: OptTensor = None, edge_weight: OptTensor = None) -> Tensor:
        """
        Encode input

        :param edge_index: Adjacency
        :param edge_weight: Edges weight (optional)
        :param input: Input graph signal (nodes × in_features)
        :param initial_state: Initial state (nodes × hidden_features) for all reservoir layers, default zeros
        :param batch: Batch index (optional)
        :return: Encoding (samples × dim)
        """
        if initial_state is None:
            initial_state = [torch.zeros(input.shape[0], layer.out_features).to(input) for layer in self.layers]
        elif len(initial_state) != self.num_layers and initial_state.dim() == 2:
            initial_state = [initial_state] * self.num_layers
        embeddings = [self._embed(self.layers[0], edge_index, edge_weight, input, initial_state[0])]
        for i in range(1, self.num_layers):
            embeddings.append(self._embed(self.layers[i], edge_index, edge_weight, embeddings[-1], initial_state[i]))
        if self.fully:
            return torch.cat([self.pooling(x, batch) if self.pooling else x for x in embeddings], dim=-1)
        else:
            return self.pooling(embeddings[-1], batch) if self.pooling else embeddings[-1]

    def _embed(self, layer: ReservoirConvLayer, edge_index: Adj, edge_weight: OptTensor, input: Tensor,
               initial_state: Tensor) -> Tensor:
        """
        Compute node embeddings for a single layer

        :param layer: Reservoir layer
        :param edge_index: Adjacency
        :param edge_weight: Edges weight (optional)
        :param input: Input graph signal (nodes × in_features)
        :param initial_state: Initial state (nodes × hidden_features) for all reservoir layers, default zeros
        :return: Encoding (nodes × dim)
        """
        iterations = 0
        old_state = initial_state
        while True:
            state = layer(edge_index, input, old_state, edge_weight)
            if self.max_iterations and iterations >= self.max_iterations:
                break
            if torch.norm(old_state - state) < self.epsilon:
                break
            old_state = state
            iterations += 1
        return state

    @property
    def out_features(self) -> int:
        """Embedding dimension"""
        return sum(layer.out_features for layer in self.layers) if self.fully else self.layers[-1].out_features


class DynamicGraphReservoir(GraphReservoir):
    """
    Reservoir for discrete-time dynamic temporal graphs

    :param num_layers: Reservoir layers
    :param in_features: Size of input
    :param hidden_features: Size of reservoir (i.e. number of hidden units per layer)
    :param bias: Whether bias term is present
    :param pooling: Graph pooling function (optional, default no pooling)
    :param fully: Whether to concatenate all layers' encodings, or use just final layer encoding
    :param return_sequences: Return the sequence of states instead of just the final states (default false)
    """
    pooling: Optional[Callable[[Tensor, Tensor], Tensor]]
    fully: bool

    def __init__(self, num_layers: int, in_features: int, hidden_features: int, bias: bool = False,
                 pooling: Optional[Callable[[Tensor, Tensor], Tensor]] = None, fully: bool = False,
                 return_sequences: bool = False, **kwargs):
        super().__init__(num_layers, in_features, hidden_features, bias, **kwargs)
        self.pooling = pooling
        self.fully = fully
        self.return_sequences = return_sequences

    def forward(self, edge_index: Union[List[Adj], Adj], input: Union[Tensor, List[Tensor]],
                initial_state: Optional[Union[List[Tensor], Tensor]] = None,
                batch: OptTensor = None, edge_weight: Optional[Union[List[Tensor], Tensor]] = None,
                mask: Optional[List[Tensor]] = None) -> Tensor:
        """
        Encode input

        :param edge_index: Sequence of adjacency matrices (time × Adj), or Adj in the case of the spatio-temporal setting
        :param edge_weight: Optional sequence of edge weights, or fixed edge weights for the spatio-temporal setting
        :param input: Input graph signal (time × nodes × in_features)
        :param initial_state: Initial state (nodes × hidden_features) for all reservoir layers, default zeros
        :param batch: Batch index (optional)
        :param mask: Sequence of node masks (optional, useful for padding dynamic graphs with different lengths)
        :return: Encoding (samples × dim)
        """
        if initial_state is None:
            num_nodes = input[0].shape[0] if isinstance(input, list) else input.shape[1]
            state = [torch.zeros(num_nodes, layer.out_features).to(input[0]) for layer in self.layers]
        elif len(initial_state) != self.num_layers and initial_state.dim() == 2:
            state = [initial_state.clone() for _ in range(self.num_layers)]
        else:
            state = [x.clone() for x in initial_state]
        if self.return_sequences:
            return self._embed_sequence(edge_index, edge_weight, input, state, mask, batch)
        else:
            return self._embed_final(edge_index, edge_weight, input, state, mask, batch)

    def _embed_final(self, edge_index: Union[List[Adj], Adj], edge_weight: Optional[Union[List[Tensor], Tensor]],
                     input: Union[Tensor, List[Tensor]], state: List[Tensor], mask: Optional[List[Tensor]],
                     batch: OptTensor) -> Tensor:
        """
        Compute final-state embedding

        :param edge_index: Sequence of adjacency matrices (time × Adj), or Adj in the case of the spatio-temporal setting
        :param edge_weight: Optional sequence of edge weights, or fixed edge weights for the spatio-temporal setting
        :param input: Input graph signal (time × nodes × in_features)
        :param state: Initial state (nodes × hidden_features) for all reservoir layers
        :param mask: Sequence of node masks (optional, useful for padding dynamic graphs with different lengths)
        :param batch: Batch index (optional)
        :return: Encoding (samples × dim)
        """
        for t in range(len(input)):
            edge_index_t = edge_index[t] if isinstance(edge_index, list) else edge_index
            edge_weight_t = edge_weight[t] if isinstance(edge_weight, list) else edge_weight
            mask_t = slice(None) if mask is None else mask[t]
            state[0][mask_t] = self.layers[0](edge_index_t, input[t], state[0], edge_weight_t)[mask_t]
            for i in range(1, self.num_layers):
                state[i][mask_t] = self.layers[i](edge_index_t, state[i - 1], state[i], edge_weight_t)[mask_t]
        if self.fully:
            return torch.cat([self.pooling(x, batch) if self.pooling else x for x in state], dim=-1)
        else:
            return self.pooling(state[-1], batch) if self.pooling else state[-1]

    def _embed_sequence(self, edge_index: Union[List[Adj], Adj], edge_weight: Optional[Union[List[Tensor], Tensor]],
                        input: Union[Tensor, List[Tensor]], state: List[Tensor], mask: Optional[List[Tensor]],
                        batch: OptTensor) -> Tensor:
        """
        Compute sequence-to-sequence embedding

        :param edge_index: Sequence of adjacency matrices (time × Adj), or Adj in the case of the spatio-temporal setting
        :param edge_weight: Optional sequence of edge weights, or fixed edge weights for the spatio-temporal setting
        :param input: Input graph signal (time × nodes × in_features)
        :param state: Initial state (nodes × hidden_features) for all reservoir layers
        :param mask: Sequence of node masks (optional, useful for padding dynamic graphs with different lengths)
        :param batch: Batch index (optional)
        :return: Encoding (time × samples × dim)
        """
        size = int(batch.max().item() + 1) if self.pooling else state[0].shape[0]
        embeddings = torch.zeros(len(input), size, self.out_features).to(input[0])
        for t in range(len(input)):
            edge_index_t = edge_index[t] if isinstance(edge_index, list) else edge_index
            edge_weight_t = edge_weight[t] if isinstance(edge_weight, list) else edge_weight
            mask_t = slice(None) if mask is None else mask[t]
            state[0][mask_t] = self.layers[0](edge_index_t, input[t], state[0], edge_weight_t)[mask_t]
            for i in range(1, self.num_layers):
                state[i][mask_t] = self.layers[i](edge_index_t, state[i - 1], state[i], edge_weight_t)[mask_t]
            if self.fully:
                embeddings[t] = torch.cat([self.pooling(x, batch) if self.pooling else x for x in state], dim=-1)
            else:
                embeddings[t] = self.pooling(state[-1], batch) if self.pooling else state[-1]
        return embeddings

    @property
    def out_features(self) -> int:
        """Embedding dimension"""
        return sum(layer.out_features for layer in self.layers) if self.fully else self.layers[-1].out_features

    def extra_repr(self) -> str:
        return f'return_sequences={self.return_sequences}'



from .data import TemporalData, DynamicData, DynamicBatch
from .readout import Readout
from .reservoir import initializer, StaticGraphReservoir, DynamicGraphReservoir

__all__ = ['StaticGraphReservoir', 'DynamicGraphReservoir', 'initializer', 'Readout', 'TemporalData', 'DynamicData',
           'DynamicBatch']



from __future__ import annotations

from collections import UserList
from typing import List, Any

from torch import Tensor
from torch_geometric.data import Data, Batch


class TemporalData(Data):
    """
    Data object describing a temporal graph.
    Attribute declared as temporal keys have first dimension as time dimension;
    second dimension assumed as node (or graph) dimension.

    :param temporal_keys: Which keys have a first temporal dimension (default only 'x')

    Example::
        >>> data = TemporalData(temporal_keys=['x', 'y'], x=torch.rand(10,7,3), edge_index=torch.tensor(...), y=torch.rand(10,1))
        # Temporal graph having 10 time-steps, 7 nodes, 3 features per node, and a sequence of graph attributes for each time-step.
        >>> Batch.from_data_list([data, data])
        TemporalDataBatch(x=[10, 14, 3], edge_index=[2, 12], y=[10, 2], batch=[14], ptr=[3])
    """

    def __init__(self, temporal_keys: List[str] = ['x'], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._temporal_keys = temporal_keys

    def __cat_dim__(self, key: str, value: Any, *args, **kwargs) -> Any:
        if key in self._temporal_keys:
            return 1
        else:
            return super().__cat_dim__(key, value, args, kwargs)

    @property
    def num_timesteps(self) -> int:
        return self[self._temporal_keys[0]].shape[0]

    @staticmethod
    def from_data(data: Data, temporal_keys: List[str] = ['x']) -> TemporalData:
        return TemporalData(temporal_keys=temporal_keys, **data.to_dict())


class DynamicData(object):
    def __init__(self, static_keys: List[str] = [], **stuff):
        self._static_keys = static_keys
        self._storage = stuff

    def __getitem__(self, t: int):
        return Data(**{key: value if key in self._static_keys else value[t] for key, value in self._storage.items()})

    def __getattr__(self, key: str):
        if key in self.keys:
            return self._storage[key]
        else:
            raise AttributeError(f"No attribute '{key}'")

    @property
    def keys(self) -> List[str]:
        return list(self._storage.keys())

    @property
    def num_timesteps(self) -> int:
        return len(self._storage[[key for key in self.keys if key not in self._static_keys][0]])

    def to(self, *args, **kwargs):
        for key in self.keys:
            if isinstance(self._storage[key], Tensor):
                self._storage[key] = self._storage[key].to(*args, **kwargs)
            else:
                for t in range(len(self._storage[key])):
                    self._storage[key][t] = self._storage[key][t].to(*args, **kwargs)
        return self

    def __repr__(self):
        return f"DynamicData({', '.join(self.keys)})"


class DynamicBatch(UserList):
    def __init__(self, data: List[DynamicData]):
        super().__init__()
        self.data = [Batch.from_data_list([sample[t] for sample in data]) for t in range(data[0].num_timesteps)]

    def to(self, *args, **kwargs):
        for t in range(len(self)):
            self.data[t] = self.data[t].to(*args, **kwargs)
        return self

    def __repr__(self):
        return f'DynamicBatch[{len(self)}]'
